// TODO: delete this file

import axios, { AxiosInstance } from 'axios'
import { API_URL } from '../config.json'
import { AuthDetails } from '../../common/types'
import log from 'electron-log'
import jwtDecode, { JwtPayload } from 'jwt-decode'

const APP_SERVICE = 'kloke-app'

type User = {
  phone: string
  accessToken: string
  tfaMissing: boolean
}

export default class AuthManager {
  private currentUser: User | null = null
  private callbacks: (() => void)[] = []
  private client: AxiosInstance

  constructor() {
    this.client = axios.create({
      baseURL: API_URL,
    })
  }

  getCurrentUser(): User | null {
    return this.currentUser
  }

  shouldRefresh(): boolean {
    if (this.currentUser === null) {
      return false
    }

    const expirationDate = AuthManager.getJwtExpirationDate(this.currentUser.accessToken)

    return expirationDate !== null && expirationDate > new Date()
  }

  async refresh(): Promise<void> {
    log.info('[Auth] Refreshing tokens')

    const refreshToken = await AuthManager.loadRefreshToken()

    if (!refreshToken) {
      return await this.logout()
    }

    try {
      const result = await this.client.post('/v1/login/refresh', {
        refreshToken,
      })

      await AuthManager.saveRefreshToken(result.data.refreshToken)

      this.currentUser = {
        phone: result.data.phone,
        accessToken: result.data.accessToken,
        tfaMissing: result.data.tfaMissing,
      }

      this.executeCallbacks()
    } catch (e) {
      log.error(`[Auth] Unable to refresh tokens: ${e.message}`)

      return await this.logout()
    }
  }

  async login(phone: string, password: string): Promise<AuthDetails> {
    log.info('[Auth] Logging in')

    try {
      const result = await this.client.post('/v1/login', {
        phone,
        password,
      })

      await AuthManager.saveRefreshToken(result.data.refreshToken)

      this.currentUser = {
        phone: result.data.phone,
        accessToken: result.data.accessToken,
        tfaMissing: result.data.tfaMissing,
      }

      this.executeCallbacks()

      return AuthManager.generateAuthDetails(this.currentUser)
    } catch (e) {
      if (e.response && e.response.status === 401) {
        throw new Error('Invalid credentials')
      }

      log.error(`[Auth] Unable to login: ${e.message}`)

      throw new Error('Unable to access Klink storage')
    }
  }

  async confirmTfa(code: string): Promise<void> {
    log.info('[Auth] Confirming TFA')

    if (this.currentUser === null) {
      throw new Error('[DEV] User must be logged before confirming TFA')
    }

    try {
      const result = await this.client.post(
        '/v1/login/tfa',
        {
          code,
        },
        {
          headers: {
            Authorization: `Bearer ${this.currentUser.accessToken}`,
          },
        }
      )

      await AuthManager.saveRefreshToken(result.data.refreshToken)

      this.currentUser = {
        phone: this.currentUser.phone,
        accessToken: result.data.accessToken,
        tfaMissing: false,
      }

      this.executeCallbacks()
    } catch (e) {
      if (e.response && e.response.status === 403) {
        throw new Error('Invalid TFA code')
      }

      log.error(`[Auth] Unable to confirm TFA: ${e.message}`)

      throw new Error('Unable to access Klink storage')
    }
  }

  async logout(): Promise<void> {
    log.info('[Auth] Logging out')

    this.currentUser = null

    await AuthManager.removeRefreshToken()

    this.executeCallbacks()
  }

  getAuthDetails(): AuthDetails | null {
    if (this.currentUser === null) {
      return null
    }

    return AuthManager.generateAuthDetails(this.currentUser)
  }

  onChange(callback: () => void): void {
    this.callbacks.push(callback)
  }

  private static async saveRefreshToken(refreshToken: string): Promise<void> {
    try {
    } catch {
      // Ignore keystore issues
    }
  }

  private static async removeRefreshToken(): Promise<void> {
    try {
    } catch {
      // Ignore keystore issues
    }
  }

  private static async loadRefreshToken(): Promise<string | null> {
    try {
      const refreshToken = localStorage.getItem(APP_SERVICE)

      if (!refreshToken) {
        return null
      }

      const expirationDate = AuthManager.getJwtExpirationDate(refreshToken)

      if (expirationDate !== null && expirationDate < new Date()) {
        return null
      }

      return refreshToken
    } catch {
      // Ignore keystore issues
      return null
    }
  }

  private static getJwtExpirationDate(jwtToken: string): Date | null {
    const { exp: expires } = jwtDecode<JwtPayload>(jwtToken)

    if (!expires) {
      return null
    }

    return new Date(expires * 1000)
  }

  private static generateAuthDetails(user: User): AuthDetails {
    return {
      phone: user.phone,
      tfaMissing: user.tfaMissing,
    }
  }

  private executeCallbacks(): void {
    this.callbacks.forEach((callback) => callback())
  }
}
