import { createHash } from 'crypto'
import fs from 'fs'
import { promisify } from 'util'
import express from 'express'
import { SignJWT } from 'jose/jwt/sign'
import { jwtVerify } from 'jose/jwt/verify'
import { generateKeyPair } from 'jose/util/generate_key_pair'
import passport from 'passport'
import { ExtractJwt, Strategy as JwtStrategy } from 'passport-jwt'
import { Strategy as LocalStrategy } from 'passport-local'
import sqlite3 from 'sqlite3'
import { v4 as uuidv4 } from 'uuid'

const db = new sqlite3.Database('./mock.sqlite')
const { publicKey, privateKey } = await generateKeyPair('ES256')

const DEFAULT_ITEMS_PER_PAGE = 10

const port = 8080
const app = express()
app.use(express.json())

const baseKlinkUrl = process.env.KLINK_APP_URL
const getKlinkUrl = (token) => `${baseKlinkUrl}/${token}`

passport.use(
  new LocalStrategy(
    {
      usernameField: 'phone',
      passwordField: 'password',
    },
    (username, password, done) => {
      const hash = createHash('sha256')

      db.get(
        'SELECT phone, tfa FROM users WHERE phone = ? AND password = ?',
        username,
        hash.update(password).digest('hex'),
        (err, res) => {
          if (err) {
            done(err)

            return
          }

          if (res) {
            done(null, res)

            return
          }

          done(null, false)
        }
      )
    }
  )
)

passport.use(
  new JwtStrategy(
    {
      jwtFromRequest: ExtractJwt.fromAuthHeaderAsBearerToken(),
      secretOrKey: publicKey.export({
        type: 'spki',
        format: 'pem',
      }),
      audience: 'access',
    },
    (payload, done) => {
      if (payload.tfaMissing) {
        done(null, false)

        return
      }

      done(null, {
        phone: payload.phone,
        tfa: payload.tfa,
      })
    }
  )
)

passport.use(
  'no-tfa-jwt',
  new JwtStrategy(
    {
      jwtFromRequest: ExtractJwt.fromAuthHeaderAsBearerToken(),
      secretOrKey: publicKey.export({
        type: 'spki',
        format: 'pem',
      }),
      audience: 'access',
    },
    (payload, done) => {
      done(null, {
        phone: payload.phone,
        tfa: payload.tfa,
      })
    }
  )
)

app.use(passport.initialize({}))

const generateTokensPair = async (user, tfaConfirmed = false) => {
  const accessToken = await new SignJWT({
    phone: user.phone,
    tfa: Boolean(user.tfa),
    tfaMissing: !tfaConfirmed,
  })
    .setProtectedHeader({ alg: 'ES256' })
    .setIssuedAt()
    .setAudience('access')
    .setExpirationTime('1h')
    .sign(privateKey)

  const refreshToken = await new SignJWT({
    phone: user.phone,
    tfa: Boolean(user.tfa),
    tfaMissing: !tfaConfirmed,
  })
    .setProtectedHeader({ alg: 'ES256' })
    .setIssuedAt()
    .setAudience('refresh')
    .setExpirationTime('1d')
    .sign(privateKey)

  return {
    accessToken,
    refreshToken,
  }
}

app.post(
  '/v1/login',
  passport.authenticate('local', {
    session: false,
  }),
  async (req, res) => {
    const tokens = await generateTokensPair(req.user, !req.user.tfa)

    res.json({
      ...tokens,
      phone: req.user.phone,
      tfaMissing: Boolean(req.user.tfa),
    })
  }
)

app.post('/v1/login/refresh', async (req, res) => {
  try {
    const { payload } = await jwtVerify(req.body.refreshToken, publicKey, {
      audience: 'refresh',
    })

    const tokens = await generateTokensPair(
      {
        phone: payload.phone,
        tfa: payload.tfa,
      },
      !payload.tfaMissing
    )

    res.json({
      ...tokens,
      phone: payload.phone,
      tfaMissing: payload.tfaMissing,
    })
  } catch (e) {
    res.json({
      done: false,
    })
  }
})

app.post(
  '/v1/login/tfa',
  passport.authenticate('no-tfa-jwt', { session: false }),
  async (req, res) => {
    if (req.body.code !== '000000') {
      res.status(403)
      res.json({
        done: false,
      })

      return
    }

    const tokens = await generateTokensPair(
      {
        phone: req.user.phone,
        tfa: req.user.tfa,
      },
      true
    )

    res.json({
      ...tokens,
      phone: req.user.phone,
    })
  }
)

app.post(
  '/v1/notifications',
  passport.authenticate('jwt', { session: false }),
  (req, res) => {
    db.serialize(() => {
      db.run(
        'UPDATE users SET device = ? WHERE phone = ?',
        req.body.deviceToken,
        req.user.phone
      )
    })

    res.json({
      done: true,
    })
  }
)

app.post('/v1/tokens', passport.authenticate('jwt', { session: false }), (req, res) => {
  const statement = db.prepare('INSERT INTO tokens VALUES (?, ?, ?, ?, ?)')

  const token = uuidv4()

  statement.run(
    token,
    req.body.title,
    req.body.data,
    req.user.phone,
    Number(req.body.enableNotifications)
  )

  const statementShareWith = db.prepare('INSERT INTO tokensShared VALUES (?, ?)')

  req.body.shareWith.forEach((phone) => {
    statementShareWith.run(token, phone)
  })

  statementShareWith.finalize()

  res.json({
    token,
  })
})

app.get(
  '/v1/tokens/:token',
  passport.authenticate('jwt', { session: false }),
  (req, res) => {
    db.get(
      'SELECT id, title, data, creator, enableNotifications FROM tokens WHERE id = ?',
      req.params.token,
      (err, dres) => {
        if (err) {
          res.status(500)
          res.send('Failed')

          return
        }

        if (!dres) {
          res.status(404)
          res.send('Not found')

          return
        }

        const result = {
          id: dres.id,
          url: getKlinkUrl(dres.id),
          editable: dres.creator === req.user.phone,
          title: dres.title,
          data: dres.data,
        }

        if (dres.creator !== req.user.phone && dres.enableNotifications) {
          db.get(
            'SELECT device FROM users WHERE phone = ?',
            dres.creator,
            (err, ddres) => {}
          )
        }

        db.all(
          'SELECT phone FROM tokensShared WHERE token = ?',
          req.params.token,
          (err, rows) => {
            if (err) {
              res.status(500)
              res.send('Failed')

              return
            }

            const shareWith = rows.map((item) => item.phone)

            if (dres.creator === req.user.phone) {
              result.enableNotifications = dres.enableNotifications
              result.shareWith = shareWith
            }

            if (dres.creator !== req.user.phone && !shareWith.includes(req.user.phone)) {
              res.status(404)
              res.send('Not found')

              return
            }

            res.json(result)
          }
        )
      }
    )
  }
)

app.get('/v1/tokens', passport.authenticate('jwt', { session: false }), (req, res) => {
  const limit = req.query.itemsPerPage || DEFAULT_ITEMS_PER_PAGE
  const page = req.query.page || 1

  db.all(
    `
        SELECT id, title, data, enableNotifications, phone, totalItems
        FROM (SELECT *, COUNT(*) OVER() AS totalItems FROM tokens WHERE creator = ? LIMIT ? OFFSET ?) userTokens 
        LEFT JOIN tokensShared ON userTokens.id = tokensShared.token 
      `,
    req.user.phone,
    limit,
    (page - 1) * limit,
    (err, rows) => {
      if (err) {
        console.log(err)
        res.status(500)
        res.send('Failed')

        return
      }

      const rowsMap = rows.reduce((map, row) => {
        if (!map.has(row.id)) {
          const rowToSave = {
            id: row.id,
            url: getKlinkUrl(row.id),
            title: row.title,
            data: row.data,
            enableNotifications: !!row.enableNotifications,
            editable: true,
            shareWith: row.phone ? [row.phone] : [],
          }
          map.set(row.id, rowToSave)
        } else if (row.phone) {
          const rowToUpdate = map.get(row.id)
          map.set(row.id, {
            ...rowToUpdate,
            shareWith: [...rowToUpdate.shareWith, row.phone],
          })
        }

        return map
      }, new Map())

      res.json({
        tokens: Array.from(rowsMap.values()),
        itemsPerPage: limit,
        totalItems: rows[0]?.totalItems || 0,
        page,
      })
    }
  )
})

app.put(
  '/v1/tokens/:token',
  passport.authenticate('jwt', { session: false }),
  (req, res) => {
    db.get('SELECT creator FROM tokens WHERE id = ?', req.params.token, (err, dres) => {
      if (!dres || dres.creator !== req.user.phone) {
        res.status(404)
        res.send('Not found')

        return
      }

      db.serialize(() => {
        const statement = db.prepare(
          'UPDATE tokens SET title = ?, data = ?, enableNotifications = ? WHERE id = ?'
        )

        statement.run(
          req.body.title,
          req.body.data,
          req.body.enableNotifications,
          req.params.token
        )

        db.run('DELETE FROM tokensShared WHERE token = ?', req.params.token)

        const statementShareWith = db.prepare('INSERT INTO tokensShared VALUES (?, ?)')

        req.body.shareWith.forEach((phone) => {
          statementShareWith.run(req.params.token, phone)
        })

        statementShareWith.finalize()
      })
    })

    res.json({
      done: true,
    })
  }
)

app.post('/v1/users', (req, res) => {
  const statement = db.prepare('INSERT INTO users VALUES (?, ?, ?)')

  const hash = createHash('sha256')

  statement.run(
    req.body.phone,
    hash.update(req.body.password).digest('hex'),
    req.body.tfa || false
  )

  res.json({
    done: true,
  })
})

const readFile = promisify(fs.readFile)

app.get('/k/:token', async (req, res) => {
  const data = await readFile('./index.html')

  const body = data.toString()

  res.header('Content-Type', 'text/html')
  res.send(body.replace(/TOKEN_PLACEHOLDER/g, req.params.token))
})

db.serialize(() => {
  db.run(
    'CREATE TABLE IF NOT EXISTS users (phone TEXT, password TEXT, tfa INTEGER, device TEXT)'
  )
  db.run(
    'CREATE TABLE IF NOT EXISTS tokens (id TEXT, title TEXT, data TEXT, creator TEXT, enableNotifications INTEGER)'
  )
  db.run('CREATE TABLE IF NOT EXISTS tokensShared (token TEXT, phone TEXT)')
})

app.listen(port, () => {
  console.log(`Mock listening at http://localhost:${port}`)
})
